# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["CircularContactPointsOp", "ContactPointsOp"]


import numpy as np

import theano
from theano import gof
import theano.tensor as tt

from .find_roots import CircularRootFinder, GeneralRootFinder


class CircularContactPointsOp(tt.Op):

    __props__ = ("tol", )
    num_inputs = 4

    def __init__(self, tol=1e-6, **kwargs):
        self.tol = float(tol)
        super(CircularContactPointsOp, self).__init__(**kwargs)

    def make_node(self, *args):
        if len(args) != self.num_inputs:
            raise ValueError("expected {0} inputs".format(self.num_input))
        dtype = theano.config.floatX
        in_args = []
        for a in args:
            try:
                a = tt.as_tensor_variable(a)
            except tt.AsTensorError:
                pass
            else:
                dtype = theano.scalar.upcast(dtype, a.dtype)
            in_args.append(a)
        ndim = in_args[0].ndim
        out_args = [
            tt.TensorType(dtype=dtype,
                          broadcastable=[False] * ndim)()
            for i in range(4)]
        return gof.Apply(self, in_args, out_args)

    def infer_shape(self, node, shapes):
        return shapes[0], shapes[0], shapes[0], shapes[0]

    def perform(self, node, inputs, outputs):
        a, i, r, R = inputs
        n_pl = a.size

        results = []
        for n in range(n_pl):
            solver = CircularRootFinder(a.flat[n], i.flat[n], tol=self.tol)
            roots = np.concatenate([solver.get_contact_points(L)
                                    for L in [R.flat[n] - r.flat[n],
                                              R.flat[n] + r.flat[n]]])
            results.append(np.sort(np.array(roots).flatten()))

        for m, roots in enumerate(zip(*results)):
            outputs[m][0] = np.array(roots).reshape(a.shape)


class ContactPointsOp(CircularContactPointsOp):
    num_inputs = 6

    def perform(self, node, inputs, outputs):
        a, e, w, i, r, R = inputs
        n_pl = a.size

        results = []
        for n in range(n_pl):
            solver = GeneralRootFinder(a.flat[n], e.flat[n], w.flat[n],
                                       i.flat[n], tol=self.tol)
            roots = np.concatenate([solver.get_contact_points(L)
                                    for L in [R.flat[n] - r.flat[n],
                                              R.flat[n] + r.flat[n]]])
            results.append(np.sort(np.array(roots).flatten()))

        for m, roots in enumerate(zip(*results)):
            outputs[m][0] = np.array(roots).reshape(a.shape)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

# import pytest
import numpy as np

import theano
import theano.tensor as tt

from . import terms
from .celerite import GP


def test_broadcast_dim():
    logS0 = tt.scalar()
    logw0 = tt.scalar()
    logQ = tt.scalar()
    kernel = terms.SHOTerm(S0=tt.exp(logS0), w0=tt.exp(logw0), Q=tt.exp(logQ))

    x = tt.vector()
    y = tt.vector()
    diag = tt.vector()
    gp = GP(kernel, x, diag, J=2)
    loglike = gp.log_likelihood(y)

    args = [logS0, logw0, logQ, x, y, diag]
    grad = theano.function(args, theano.grad(loglike, args))

    np.random.seed(42)
    N = 50
    x = np.sort(10 * np.random.rand(N))
    y = np.sin(x)
    diag = np.random.rand(N)
    grad(-5.0, -2.0, 1.0, x, y, diag)


def _get_theano_kernel(celerite_kernel):
    import celerite.terms as cterms
    if isinstance(celerite_kernel, cterms.TermSum):
        result = _get_theano_kernel(celerite_kernel.terms[0])
        for k in celerite_kernel.terms[1:]:
            result += _get_theano_kernel(k)
        return result
    elif isinstance(celerite_kernel, cterms.TermProduct):
        return (
            _get_theano_kernel(celerite_kernel.k1) *
            _get_theano_kernel(celerite_kernel.k2))
    elif isinstance(celerite_kernel, cterms.RealTerm):
        return terms.RealTerm(log_a=celerite_kernel.log_a,
                              log_c=celerite_kernel.log_c)
    elif isinstance(celerite_kernel, cterms.ComplexTerm):
        if not celerite_kernel.fit_b:
            return terms.ComplexTerm(log_a=celerite_kernel.log_a,
                                     b=0.0,
                                     log_c=celerite_kernel.log_c,
                                     log_d=celerite_kernel.log_d)
        return terms.ComplexTerm(log_a=celerite_kernel.log_a,
                                 log_b=celerite_kernel.log_b,
                                 log_c=celerite_kernel.log_c,
                                 log_d=celerite_kernel.log_d)
    elif isinstance(celerite_kernel, cterms.SHOTerm):
        return terms.SHOTerm(log_S0=celerite_kernel.log_S0,
                             log_Q=celerite_kernel.log_Q,
                             log_w0=celerite_kernel.log_omega0)
    elif isinstance(celerite_kernel, cterms.Matern32Term):
        return terms.Matern32Term(log_sigma=celerite_kernel.log_sigma,
                                  log_rho=celerite_kernel.log_rho)
    raise NotImplementedError()


# @pytest.mark.parametrize(
#     "celerite_kernel",
#     [
#         "cterms.RealTerm(log_a=0.1, log_c=0.5)",
#         "cterms.RealTerm(log_a=0.1, log_c=0.5) + "
#         "cterms.RealTerm(log_a=-0.1, log_c=0.7)",
#         "cterms.ComplexTerm(log_a=0.1, log_c=0.5, log_d=0.1)",
#         "cterms.ComplexTerm(log_a=0.1, log_b=-0.2, log_c=0.5, log_d=0.1)",
#         "cterms.SHOTerm(log_S0=0.1, log_Q=-1, log_omega0=0.5)",
#         "cterms.SHOTerm(log_S0=0.1, log_Q=1.0, log_omega0=0.5)",
#         "cterms.SHOTerm(log_S0=0.1, log_Q=1.0, log_omega0=0.5) + "
#         "cterms.RealTerm(log_a=0.1, log_c=0.4)",
#         "cterms.SHOTerm(log_S0=0.1, log_Q=1.0, log_omega0=0.5) * "
#         "cterms.RealTerm(log_a=0.1, log_c=0.4)",
#         "cterms.Matern32Term(log_sigma=0.1, log_rho=0.4)",
#     ]
# )
# def test_gp(celerite_kernel, seed=1234):
#     import celerite
#     import celerite.terms as cterms  # NOQA
#     celerite_kernel = eval(celerite_kernel)
#     np.random.seed(seed)
#     x = np.sort(np.random.rand(100))
#     yerr = np.random.uniform(0.1, 0.5, len(x))
#     y = np.sin(x)
#     diag = yerr**2

#     celerite_gp = celerite.GP(celerite_kernel)
#     celerite_gp.compute(x, yerr)
#     celerite_loglike = celerite_gp.log_likelihood(y)

#     kernel = _get_theano_kernel(celerite_kernel)
#     gp = GP(kernel, x, diag)
#     loglike = gp.log_likelihood(y).eval()

#     assert np.allclose(loglike, celerite_loglike)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["FactorRevOp"]

from .base_op import CeleriteBaseOp


class FactorRevOp(CeleriteBaseOp):

    func_file = "./factor_rev.cc"
    func_name = "APPLY_SPECIFIC(factor_rev)"
    num_input = 7
    output_ndim = (1, 2, 2, 2)

    def __init__(self, J=-1):
        super(FactorRevOp, self).__init__(J=J)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["CeleriteBaseOp"]

import pkg_resources

import theano
from theano import gof
import theano.tensor as tt

from ..build_utils import get_compile_args, get_cache_version


class CeleriteBaseOp(gof.COp):

    __props__ = ("J", "n_rhs")
    num_input = 0
    output_ndim = ()
    func_file = None
    func_name = None

    def __init__(self, J=-1, n_rhs=-1):
        self.J = int(J)
        self.n_rhs = int(n_rhs)
        super(CeleriteBaseOp, self).__init__(self.func_file, self.func_name)

    def c_code_cache_version(self):
        return get_cache_version()

    def c_headers(self, compiler):
        return ["theano_helpers.h"]

    def c_header_dirs(self, compiler):
        return [
            pkg_resources.resource_filename(__name__, "include"),
            pkg_resources.resource_filename("exoplanet.theano_ops.starry",
                                            "starry/lib/eigen_3.3.3")
        ]

    def c_compile_args(self, compiler):
        args = get_compile_args(compiler)
        if 0 < self.J <= 32:
            args.append("-DCELERITE_J={0}".format(self.J))
            args.append("-DCELERITE_J2={0}".format(self.J * self.J))
            if self.J == 1:
                args.append("-DCELERITE_J_ORDER=Eigen::ColMajor")
            else:
                args.append("-DCELERITE_J_ORDER=Eigen::RowMajor")
        if 0 < self.n_rhs <= 32:
            args.append("-DCELERITE_NRHS={0}".format(self.n_rhs))
            args.append("-DCELERITE_JNRHS={0}".format(self.n_rhs * self.J))
            if self.n_rhs == 1:
                args.append("-DCELERITE_NRHS_ORDER=Eigen::ColMajor")
            else:
                args.append("-DCELERITE_NRHS_ORDER=Eigen::RowMajor")
            if self.n_rhs == 1 and self.J == 1:
                args.append("-DCELERITE_JNRHS_ORDER=Eigen::ColMajor")
            else:
                args.append("-DCELERITE_JNRHS_ORDER=Eigen::RowMajor")
        return args

    def _get_node_args(self, *args):
        if len(args) != self.num_input:
            raise ValueError("expected {0} inputs".format(self.num_input))
        dtype = theano.config.floatX
        in_args = []
        for a in args:
            try:
                a = tt.as_tensor_variable(a)
            except tt.AsTensorError:
                pass
            else:
                dtype = theano.scalar.upcast(dtype, a.dtype)
            in_args.append(a)
        out_args = [
            tt.TensorType(dtype=dtype, broadcastable=[False] * ndim)()
            for ndim in self.output_ndim]
        return in_args, out_args

    def make_node(self, *args):
        return gof.Apply(self, *self._get_node_args(*args))



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["GeneralRootFinder", "CircularRootFinder"]

import numpy as np


class GeneralRootFinder(object):

    def __init__(self, semimajor, ecc, omega, incl, tol=1e-6):
        self.ecc = float(ecc)
        self.omega = float(omega)
        self.tol = float(tol)
        self.semimajor = float(semimajor)
        self.incl = float(incl)
        self.sini = np.sin(self.incl)
        self.cosi = np.cos(self.incl)
        self.cos2i = self.cosi * self.cosi
        self.setup_poly()

    def setup_poly(self):
        self.Efactor = np.sqrt((1 - self.ecc) / (1 + self.ecc))
        self.ome2 = 1 - self.ecc**2
        self.sinw = np.sin(self.omega)
        self.cosw = np.cos(self.omega)
        self.sin2w = self.sinw*self.sinw
        self.cos2w = self.cosw*self.cosw
        self.f0 = 2 * np.arctan2(self.cosw, 1 + self.sinw)

        self.a = 1.0
        self.b = 0.0
        self.c = 1 / self.ome2
        self.d = -2*self.ecc
        self.e = 0.0
        self.f = -self.ome2

        self.A = self.cos2i*self.sin2w + self.cos2w
        self.B = 2*self.cosw*self.sinw*(self.cos2i - 1)
        self.C = self.cos2i*self.cos2w + self.sin2w
        self.D = 0.0
        self.E = 0.0

    def setup_quad(self, L):
        F = -(L / self.semimajor)**2

        c, d, f = self.c, self.d, self.f
        A, B, C = self.A, self.B, self.C

        # Warning: special case for constant a, b, e, D, and E
        a0 = C**2*f**2 - 2*C*F*c*f + F**2*c**2
        a1 = 2*C*d*(C*f - F*c)
        a2 = -2*A*C*c*f+2*A*F*c**2+B**2*c*f+C**2*d**2 + 2*C**2*f - 2*C*F*c
        a3 = d*(-2*A*C*c + B**2*c + 2*C**2)
        a4 = A**2*c**2 - 2*A*C*c + B**2*c + C**2

        return a0, a1, a2, a3, a4

    def _get_all_roots(self, L):
        a0, a1, a2, a3, a4 = self. setup_quad(L)

        comp = np.eye(4, k=-1)
        comp[0, -1] = -a0 / a4
        comp[1, -1] = -a1 / a4
        comp[2, -1] = -a2 / a4
        comp[3, -1] = -a3 / a4
        roots = np.linalg.eigvals(comp)

        return roots

    def get_poly(self, L):
        return np.polynomial.Polynomial(self.setup_quad(L))

    def get_real_roots(self, L):
        all_roots = self._get_all_roots(L)

        # Only select real roots
        roots = np.real(all_roots[np.abs(np.imag(all_roots)) < self.tol])

        angles = []
        for x in roots:
            y2 = (1 - (x - self.ecc)**2) * self.ome2
            if y2 < 0.0:
                continue

            absy = np.sqrt(y2)
            for sgn in [-1, 1]:
                y = sgn * absy
                x2 = self.cosw*x - self.sinw*y
                y2 = self.cosi*(self.cosw*y + self.sinw*x)
                z2 = self.sini*(self.cosw*y + self.sinw*x)

                b2 = x2*x2 + y2*y2
                if np.abs(b2 - (L / self.semimajor)**2) < self.tol and z2 < 0:
                    angle = np.arctan2(y, x) - np.pi
                    if angle < -np.pi:
                        angle += 2*np.pi
                    angles.append(angle - self.f0)

        return angles

    def get_unique_and_wrapped_roots(self, L):
        angles = self.get_real_roots(L)

        if len(angles) < 2:
            angles = np.array([-np.pi, np.pi])
        else:
            # Deal with multiple roots
            angles = np.array(angles)
            angles.sort()
            delta = np.append(True, np.diff(angles))
            angles = angles[delta > self.tol]

            # Wrap the roots properly to span the transit
            if len(angles) == 2:
                if np.all(angles > 0):
                    angles = np.array([angles[1] - 2*np.pi, angles[0]])
                if np.all(angles < 0):
                    angles = np.array([angles[1], angles[0] + 2*np.pi])
            else:
                angles = np.array([-np.pi, np.pi])

        return angles

    def get_contact_points(self, L):
        angles = self.get_unique_and_wrapped_roots(L)

        # Convert to mean anomaly
        E = 2 * np.arctan(self.Efactor * np.tan(0.5*(angles + self.f0)))
        M = E - self.ecc * np.sin(E)

        return M


class CircularRootFinder(GeneralRootFinder):

    def __init__(self, semimajor, incl, tol=1e-6):
        super(CircularRootFinder, self).__init__(semimajor, 0.0, 0.0, incl,
                                                 tol=tol)

    def setup_poly(self):
        self.Efactor = 1.0
        self.ome2 = 1.0
        self.sinw = 0.0
        self.cosw = 1.0
        self.sin2w = 0.0
        self.cos2w = 1.0
        self.f0 = 0.5 * np.pi

        self.a = 1.0
        self.b = 0.0
        self.c = 1.0
        self.d = 0.0
        self.e = 0.0
        self.f = -1.0

        self.A = 1.0
        self.B = 0.0
        self.C = self.cosi**2
        self.D = 0.0
        self.E = 0.0

    def setup_quad(self, L):
        F = -(L / self.semimajor)**2
        C = self.C

        # Warning: special case for ecc = 0 and omega = 0
        a0 = C**2 + 2*C*F + F**2
        a1 = 0.0
        a2 = -2*C**2 - 2*C*F + 2*C + 2*F
        a3 = 0.0
        a4 = C**2 - 2*C + 1
        return a0, a1, a2, a3, a4

    def _get_all_roots(self, L):
        a0, a1, a2, a3, a4 = self.setup_quad(L)
        denom = 0.5 / a4
        first = -a2 * denom
        second = np.sqrt(a2 * a2 - 4 * a0 * a4 + 0j) * denom
        roots = np.sqrt([first + second, first - second, 0, 0])
        roots[2:] = -roots[:2]
        return roots



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["DiagDotOp"]

import theano.tensor as tt

from .base_op import CeleriteBaseOp


class DiagDotOp(CeleriteBaseOp):

    func_file = "./diag_dot.cc"
    func_name = "APPLY_SPECIFIC(diag_dot)"
    num_input = 2
    output_ndim = (1, )

    def grad(self, inputs, gradients):
        A, B = inputs
        return (
            gradients[0][:, None] * tt.transpose(B),
            gradients[1][None, :] * tt.transpose(A)
        )



# -*- coding: utf-8 -*-

from __future__ import division, print_function

import numpy as np

import theano
import theano.tensor as tt
from theano.tests import unittest_tools as utt

import starry

from .orbits import KeplerianOrbit
from .light_curves import StarryLightCurve


def test_light_curve():
    u = tt.vector()
    b = tt.vector()
    r = tt.vector()
    lc = StarryLightCurve(u)
    f = lc._compute_light_curve(b, r)
    func = theano.function([u, b, r], f)

    u_val = np.array([0.2, 0.3, 0.1, 0.5])
    b_val = np.linspace(-1.5, 1.5, 100)
    r_val = 0.1 + np.zeros_like(b_val)

    m = starry.Map(lmax=len(u_val))
    m[:] = u_val
    expect = m.flux(xo=b_val, ro=r_val) - 1

    evaluated = func(u_val, b_val, r_val)

    utt.assert_allclose(expect, evaluated)


def test_light_curve_grad():
    u_val = np.array([0.2, 0.3, 0.1, 0.5])
    b_val = np.linspace(-1.5, 1.5, 20)
    r_val = 0.1 + np.zeros_like(b_val)

    lc = lambda u, b, r: StarryLightCurve(u)._compute_light_curve(b, r)  # NOQA
    utt.verify_grad(lc, [u_val, b_val, r_val])


def test_in_transit():
    t = np.linspace(-20, 20, 1000)
    m_planet = np.array([0.3, 0.5])
    m_star = 1.45
    orbit = KeplerianOrbit(
        m_star=m_star,
        r_star=1.5,
        t0=np.array([0.5, 17.4]),
        period=np.array([10.0, 5.3]),
        ecc=np.array([0.1, 0.8]),
        omega=np.array([0.5, 1.3]),
        m_planet=m_planet,
    )
    u = np.array([0.2, 0.3, 0.1, 0.5])
    r = np.array([0.1, 0.01])

    lc = StarryLightCurve(u)
    model1 = lc.get_light_curve(r=r, orbit=orbit, t=t)
    model2 = lc.get_light_curve(r=r, orbit=orbit, t=t, use_in_transit=False)
    vals = theano.function([], [model1, model2])()
    utt.assert_allclose(*vals)

    model1 = lc.get_light_curve(r=r, orbit=orbit, t=t, texp=0.1)
    model2 = lc.get_light_curve(r=r, orbit=orbit, t=t, texp=0.1,
                                use_in_transit=False)
    vals = theano.function([], [model1, model2])()
    utt.assert_allclose(*vals)


def test_small_star():
    from batman.transitmodel import TransitModel, TransitParams
    u_star = [0.2, 0.1]
    r = 0.04221468

    m_star = 0.151
    r_star = 0.189
    period = 0.4626413
    t0 = 0.2
    b = 0.5
    ecc = 0.1
    omega = 0.1
    t = np.linspace(0, period, 500)

    r_pl = r * r_star

    orbit = KeplerianOrbit(
        r_star=r_star, m_star=m_star,
        period=period, t0=t0, b=b,
        ecc=ecc, omega=omega)
    a = orbit.a.eval()
    incl = orbit.incl.eval()

    lc = StarryLightCurve(u_star)

    model1 = lc.get_light_curve(r=r_pl, orbit=orbit, t=t)
    model2 = lc.get_light_curve(r=r_pl, orbit=orbit, t=t, use_in_transit=False)
    vals = theano.function([], [model1, model2])()
    utt.assert_allclose(*vals)

    params = TransitParams()
    params.t0 = t0
    params.per = period
    params.rp = r
    params.a = a / r_star
    params.inc = np.degrees(incl)
    params.ecc = ecc
    params.w = np.degrees(omega)
    params.u = u_star
    params.limb_dark = "quadratic"

    model = TransitModel(params, t)
    flux = model.light_curve(params)
    utt.assert_allclose(vals[0][:, 0], flux - 1)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["KeplerOp"]

import pkg_resources

import theano
from theano import gof
import theano.tensor as tt

from ..build_utils import get_cache_version


class KeplerOp(gof.COp):
    params_type = gof.ParamsType(
        maxiter=theano.scalar.int64,
        tol=theano.scalar.float64,
    )
    __props__ = ("tol", "maxiter")
    func_file = "./solver.cc"
    func_name = "APPLY_SPECIFIC(solver)"

    def __init__(self, tol=1e-12, maxiter=1000, **kwargs):
        self.tol = float(tol)
        self.maxiter = int(maxiter)
        super(KeplerOp, self).__init__(self.func_file, self.func_name)

    def c_code_cache_version(self):
        return get_cache_version()

    def c_headers(self, compiler):
        return ["theano_helpers.h", "solver.h"]

    def c_header_dirs(self, compiler):
        return [pkg_resources.resource_filename(__name__, "include")]

    def make_node(self, mean_anom, eccen):
        in_args = [tt.as_tensor_variable(mean_anom),
                   tt.as_tensor_variable(eccen)]
        return gof.Apply(self, in_args, [in_args[0].type(), in_args[0].type()])

    def infer_shape(self, node, shapes):
        return shapes[0], shapes[0]

    def grad(self, inputs, gradients):
        M, e = inputs
        E, f = self(M, e)

        bM = tt.zeros_like(M)
        be = tt.zeros_like(M)
        ecosE = e * tt.cos(E)

        if not isinstance(gradients[0].type, theano.gradient.DisconnectedType):
            # Backpropagate E_bar
            bM = gradients[0] / (1 - ecosE)
            be = tt.sin(E) * bM

        if not isinstance(gradients[1].type, theano.gradient.DisconnectedType):
            # Backpropagate f_bar
            sinf2 = tt.sin(0.5*f)
            cosf2 = tt.cos(0.5*f)
            tanf2 = sinf2 / cosf2
            e2 = e**2
            ome2 = 1 - e2
            ome = 1 - e
            ope = 1 + e
            cosf22 = cosf2**2
            twoecosf22 = 2 * e * cosf22
            factor = tt.sqrt(ope/ome)
            inner = (twoecosf22+ome) * tt.as_tensor_variable(gradients[1])

            bM += factor*(ome*tanf2**2+ope)*inner*cosf22/(ope*ome2)
            be += -2*cosf22*tanf2/ome2**2*inner*(ecosE-2+e2)

        return [bM, be]

    def R_op(self, inputs, eval_points):
        if eval_points[0] is None:
            return eval_points
        return self.grad(inputs, eval_points)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

import numpy as np

import theano
import theano.tensor as tt
from theano.tests import unittest_tools as utt

from .contact_points import ContactPointsOp, CircularContactPointsOp


class TestContactPoints(utt.InferShapeTester):

    def setUp(self):
        super(TestContactPoints, self).setUp()
        self.op_class = ContactPointsOp
        self.op = ContactPointsOp()

    def test_infer_shape(self):
        np.random.seed(42)
        args = [tt.dvector() for i in range(6)]
        vals = [np.random.rand(50) for i in range(6)]
        self._compile_and_check(args,
                                self.op(*args),
                                vals,
                                self.op_class)

    def test_basic(self):
        a = np.float64(100.0)
        e = np.float64(0.3)
        w = np.float64(0.1)
        i = np.float64(0.5*np.pi)
        r = np.float64(0.1)
        R = np.float64(1.1)

        M_expect = np.array([0.88452506, 0.8863776, 0.90490204, 0.90675455])
        M_calc = theano.function([], self.op(a, e, w, i, r, R))()

        utt.assert_allclose(M_calc, M_expect)


class TestCircularContactPoints(utt.InferShapeTester):

    def setUp(self):
        super(TestCircularContactPoints, self).setUp()
        self.op_class = CircularContactPointsOp
        self.op = CircularContactPointsOp()

    def test_infer_shape(self):
        np.random.seed(42)
        args = [tt.dvector() for i in range(4)]
        vals = [np.random.rand(50) for i in range(4)]
        self._compile_and_check(args,
                                self.op(*args),
                                vals,
                                self.op_class)

    def test_basic(self):
        a = np.float64(10.0)
        e = np.float64(0.0)
        w = np.float64(0.0)
        i = np.float64(0.5*np.pi)
        r = np.float64(0.1)
        R = np.float64(1.1)

        M_circ = theano.function([], self.op(a, i, r, R))()
        M_gen = theano.function([], ContactPointsOp()(a, e, w, i, r, R))()

        utt.assert_allclose(M_circ, M_gen)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["StarryBaseOp"]

import pkg_resources

from theano import gof

from ..build_utils import get_compile_args, get_cache_version


class StarryBaseOp(gof.COp):

    __props__ = ()
    func_file = None
    func_name = None

    def __init__(self):
        super(StarryBaseOp, self).__init__(self.func_file, self.func_name)

    def c_code_cache_version(self):
        return get_cache_version()

    def c_headers(self, compiler):
        return ["theano_helpers.h"]

    def c_header_dirs(self, compiler):
        return [
            pkg_resources.resource_filename(__name__, "include"),
            pkg_resources.resource_filename(__name__, "starry/starry"),
            pkg_resources.resource_filename(__name__,
                                            "starry/lib/eigen_3.3.3"),
            pkg_resources.resource_filename(__name__,
                                            "starry/lib/boost_1_66_0"),
        ]

    def c_compile_args(self, compiler):
        return get_compile_args(compiler)



# -*- coding: utf-8 -*-

from __future__ import division, print_function

__all__ = ["SolveRevOp"]

from .base_op import CeleriteBaseOp


class SolveRevOp(CeleriteBaseOp):

    func_file = "./solve_rev.cc"
    func_name = "APPLY_SPECIFIC(solve_rev)"
    num_input = 8
    output_ndim = (2, 2, 1, 2, 2)
